use std::borrow::Cow;

use maybe_owned::MaybeOwned;
use url::Url;

use crate::{
    common::*,
    response::{self, OffsetConnection},
};

type Parameter<'a> = (&'a str, &'a str);

pub struct RequestBuilder<'a> {
    client: Option<&'a reqwest::Client>,
    api_version: APIVersion,
    limit: u8,
    offset: usize,
    order: Order,
    stream: Option<Stream>,
    platform: Option<Platform>,
    architecture: Option<Architecture>,
    version: Option<Version>,
}

impl Default for RequestBuilder<'_> {
    fn default() -> Self {
        Self {
            limit: 10,
            client: Default::default(),
            api_version: Default::default(),
            offset: Default::default(),
            order: Default::default(),
            stream: Default::default(),
            platform: Default::default(),
            architecture: Default::default(),
            version: Default::default(),
        }
    }
}

impl<'a> RequestBuilder<'a> {
    pub fn new(client: &'a reqwest::Client) -> Self {
        Self {
            client: Some(client),
            ..Default::default()
        }
    }

    pub fn with_limit(self, limit: u8) -> Self {
        Self { limit, ..self }
    }

    pub fn with_offset(self, offset: usize) -> Self {
        Self { offset, ..self }
    }

    pub fn with_order(self, order: Order) -> Self {
        Self { order, ..self }
    }

    pub fn with_stream(self, stream: Stream) -> Self {
        Self {
            stream: Some(stream),
            ..self
        }
    }

    pub fn with_platform(self, platform: Platform) -> Self {
        Self {
            platform: Some(platform),
            ..self
        }
    }

    pub fn with_architecture(self, architecture: Architecture) -> Self {
        Self {
            architecture: Some(architecture),
            ..self
        }
    }

    pub fn with_version(self, version: String) -> Self {
        Self {
            version: Some(Version(version)),
            ..self
        }
    }

    pub async fn send(self) -> Result<OffsetConnection, response::Error> {
        Into::<Request>::into(self).send().await
    }
}

impl<'a> From<RequestBuilder<'a>> for Request<'a> {
    fn from(val: RequestBuilder<'a>) -> Self {
        Request {
            client: Cow::Borrowed(val.client.unwrap()),
            api_version: val.api_version.to_string(),
            limit: val.limit.to_string(),
            offset: val.offset.to_string(),
            order: val.order.to_string(),
            stream: val.stream.map_or_else(String::default, |s| s.to_string()),

            platform: val
                .platform
                .map_or_else(String::default, |p| p.to_string()),

            architecture: val
                .architecture
                .map_or_else(String::default, |a| a.to_string()),

            version: val
                .version
                .map_or_else(String::default, |v| v.0.to_string()),
        }
    }
}

pub struct Request<'a> {
    client: Cow<'a, reqwest::Client>,
    api_version: String,
    limit: String,
    offset: String,
    order: String,
    stream: String,
    platform: String,
    architecture: String,
    version: String,
}

impl Request<'_> {
    const API_URI: &'static str = "https://services.api.unity.com/unity/editor/release/";

    fn as_parameters(&self) -> impl Iterator<Item = Parameter> {
        [
            ("limit", self.limit.as_str()),
            ("offset", self.offset.as_str()),
            ("order", self.order.as_str()),
            ("stream", self.stream.as_str()),
            ("platform", self.platform.as_str()),
            ("architecture", self.architecture.as_str()),
            ("version", self.version.as_str()),
        ]
        .into_iter()
        .filter(|p| !p.1.is_empty())
    }

    fn as_dir(value: &str) -> String {
        format!("{value}/")
    }

    fn api_root(&self) -> Result<Url, url::ParseError> {
        Url::parse(Self::API_URI)
            .and_then(|url| url.join(&Self::as_dir(&self.api_version)))
            .and_then(|url| url.join("releases"))
    }

    fn build_uri(&self) -> Result<Url, url::ParseError> {
        self.api_root()
            .and_then(|url| Url::parse_with_params(url.as_str(), self.as_parameters()))
    }

    pub async fn send(&self) -> Result<OffsetConnection, response::Error> {
        let response = self.client.get(self.build_uri().unwrap()).send().await?;
        response::parse(response).await
    }
}

#[derive(Default)]
pub struct UnityReleaseClient<'a> {
    client: MaybeOwned<'a, reqwest::Client>,
}

impl<'a> UnityReleaseClient<'a> {
    pub fn new(client: impl Into<MaybeOwned<'a, reqwest::Client>>) -> Self {
        Self {
            client: client.into(),
        }
    }

    pub fn request(&self) -> RequestBuilder {
        match &self.client {
            MaybeOwned::Owned(x) => RequestBuilder::new(x),
            MaybeOwned::Borrowed(x) => RequestBuilder::new(x),
        }
    }

    pub async fn send(
        &self,
        request: impl Into<Request<'a>>,
    ) -> Result<OffsetConnection, response::Error> {
        request.into().send().await
    }
}

#[cfg(test)]
mod tests {
    use crate::request::*;

    #[test]
    fn defaut_request_uri() {
        let expected = "https://services.api.unity.com/unity/editor/release/v1/releases?limit=10&offset=0&order=RELEASE_DATE_DESC";
        let client = UnityReleaseClient::default();
        let request: Request = client.request().into();
        assert_eq!(request.build_uri().unwrap().as_str(), expected);
    }

    #[test]
    fn custom_request_uri() {
        let expected = "https://services.api.unity.com/unity/editor/release/v1/releases?limit=5&offset=20&order=RELEASE_DATE_ASC&stream=LTS&platform=LINUX&architecture=ARM64&version=2020.3.44f1";

        let client = UnityReleaseClient::default();
        let request: Request = client
            .request()
            .with_order(Order::Ascending)
            .with_limit(5)
            .with_offset(20)
            .with_stream(Stream::LTS)
            .with_platform(Platform::Linux)
            .with_architecture(Architecture::Arm64)
            .with_version("2020.3.44f1".to_string())
            .into();

        assert_eq!(request.build_uri().unwrap().as_str(), expected);
    }

    #[test]
    fn borrowed_client() {
        let expected = "https://services.api.unity.com/unity/editor/release/v1/releases?limit=10&offset=0&order=RELEASE_DATE_ASC";
        let request_client = reqwest::Client::default();
        let client = UnityReleaseClient::new(request_client);
        let request: Request = client.request().with_order(Order::Ascending).into();

        assert_eq!(request.build_uri().unwrap().as_str(), expected);
    }

    // #[tokio::test]
    // async fn test_connection() {
    //     let client = UnityReleaseClient::default();
    //     let request = client.request().with_offset(25).with_limit(25);
    //     let req: Request = request.into();
    //     // println!("{}", req.build_uri().unwrap());
    //     let response = req.send().await;
    //     print!("{response:?}");
    // }
}
