use std::path::PathBuf;

use reqwest::Response;
use serde::Deserialize;
use subresource_integrity::Integrity;
use time::Date;

use crate::common::*;

fn handle_fractional_numbers<'de, D>(deserializer: D) -> Result<usize, D::Error>
where
    D: serde::Deserializer<'de>,
{
    let value: f64 = Deserialize::deserialize(deserializer)?;
    let value = value.ceil();
    Ok(value as usize)
}

#[derive(Debug, serde::Deserialize)]
#[serde(rename_all = "SCREAMING_SNAKE_CASE")]
pub enum UnitKind {
    Byte,
    Kilobyte,
    Megabyte,
    Gigabyte,
}

#[derive(Debug, serde::Deserialize)]
pub struct DigitalValue {
    #[serde(deserialize_with = "handle_fractional_numbers")]
    pub value: usize,
    pub unit: UnitKind,
}

#[derive(Debug, serde::Deserialize)]
pub struct ReleaseNotes {
    pub url: PathBuf,
    pub integrity: Option<Integrity>,
    #[serde(rename = "type")]
    pub kind: FileKind,
}

#[derive(Debug, serde::Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct ThirdPartyNotice {
    #[serde(rename = "originalFileName")]
    pub original_filename: String,
    pub url: PathBuf,
    pub integrity: Option<Integrity>,
    #[serde(rename = "type")]
    pub kind: FileKind,
}

#[derive(Debug, serde::Deserialize)]
#[serde(rename_all = "UPPERCASE")]
pub enum FileKind {
    Text,
    #[serde(rename = "TAR_GZ")]
    TarGZ,
    #[serde(rename = "TAR_XZ")]
    TarXZ,
    ZIP,
    PKG,
    EXE,
    PO,
    DMG,
    LZMA,
    LZ4,
    MD,
    PDF,
}

#[derive(Debug, serde::Deserialize)]
#[serde(rename_all = "SCREAMING_SNAKE_CASE")]
pub enum ModuleCategory {
    Documentation,
    Platform,
    LanguagePack,
    DevTool,
    Plugin,
    Component,
}

#[derive(Debug, serde::Deserialize)]
#[serde(rename_all = "SCREAMING_SNAKE_CASE")]
pub enum SKUFamily {
    DOTS,
    Classic,
}

#[derive(Debug, serde::Deserialize)]
pub struct ExtractedPathRename {
    pub from: PathBuf,
    pub to: PathBuf,
}

#[derive(Debug, serde::Deserialize)]
pub struct Eula {
    pub url: PathBuf,
    pub integrity: Option<Integrity>,
    #[serde(rename = "type")]
    pub kind: FileKind,
    pub label: String,
    pub message: String,
}

#[derive(Debug, serde::Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct Module {
    pub id: String,
    pub slug: String,
    pub name: String,
    pub description: String,
    pub category: ModuleCategory,
    pub url: PathBuf,
    #[serde(rename = "type")]
    pub kind: FileKind,
    pub download_size: DigitalValue,
    pub installed_size: DigitalValue,
    pub sub_modules: Vec<Module>,
    pub required: bool,
    pub hidden: bool,
    pub extracted_path_rename: Option<ExtractedPathRename>,
    #[serde(rename = "preSelected")]
    pub preselected: bool,
    pub destination: Option<String>,
    pub eula: Option<Vec<Eula>>,
}

#[derive(Debug, serde::Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct Download {
    pub url: PathBuf,
    pub integrity: Option<Integrity>,
    #[serde(rename = "type")]
    pub kind: FileKind,
    pub platform: Platform,
    pub architecture: Architecture,
    pub download_size: DigitalValue,
    pub installed_size: DigitalValue,
    pub modules: Vec<Module>,
}

#[derive(Debug, serde::Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct Release {
    pub version: Version,
    pub date: Option<Date>,
    pub notes: Option<ReleaseNotes>,
    pub stream: Stream,
    pub downloads: Vec<Download>,
    pub sku_family: SKUFamily,
    pub recommended: bool,
    pub unity_hub_link: Option<String>,
    pub short_revision: String,
    pub third_party_notices: Vec<ThirdPartyNotice>,
}

#[derive(Debug, serde::Deserialize)]
pub struct OffsetConnection {
    pub offset: usize,
    pub limit: u8,
    pub total: usize,
    pub results: Vec<Release>,
}

pub async fn parse(response: Response) -> Result<OffsetConnection, Error> {
    let body = response.text().await?;
    println!("{body}");
    serde_json::from_str::<ResponseKind>(body.as_str())
        .map(Result::from)
        .map_err(Error::from)
        .flatten()
}

#[derive(Debug, Deserialize)]
#[serde(untagged)]
pub enum ResponseKind {
    Ok(OffsetConnection),
    Err(APIError),
}

#[derive(Debug)]
pub enum Error {
    API(APIError),
    Parse(serde_json::Error),
    Request(reqwest::Error),
}

impl From<ResponseKind> for Result<OffsetConnection, Error> {
    fn from(value: ResponseKind) -> Self {
        match value {
            ResponseKind::Ok(v) => Ok(v),
            ResponseKind::Err(e) => Err(Error::from(e)),
        }
    }
}

impl From<APIError> for Error {
    fn from(value: APIError) -> Self {
        Self::API(value)
    }
}

impl From<serde_json::Error> for Error {
    fn from(value: serde_json::Error) -> Self {
        Self::Parse(value)
    }
}

impl From<reqwest::Error> for Error {
    fn from(value: reqwest::Error) -> Self {
        Self::Request(value)
    }
}

#[derive(Debug, serde::Deserialize)]
pub struct APIError {
    pub title: String,
    pub status: u16,
    pub detail: String,
}
